/*
 * main.c
 *
 *  Created on: Aug 9, 2017
 *      Author: Christopher Hunt <chunt11@jhmi.edu>
 */

#define M_DEBUG								0
#define NRF_ESB_CHANNEL						64
#define RX_PAYLOAD_LENGTH					52
#define MAX_PERIPHERALS						8
#define MAX_UART_PACKET_SIZE				128
											// 416 for quat, accel, gyro, mag
#include "common.h"
#include "nrf_esb.h"
#include "nrf_drv_uart.h"

static bool active[ MAX_PERIPHERALS ];
static bool received[ MAX_PERIPHERALS ];

static nrf_esb_payload_t rx_payload[ MAX_PERIPHERALS ];
static uint8_t tx_uart_pkt[ MAX_UART_PACKET_SIZE ];
static nrf_drv_uart_t uart = NRF_DRV_UART_INSTANCE( UART0_INSTANCE_INDEX );

#if M_DEBUG
	static char DEBUG_MSG[ 32 ];
#endif

/* Event handler for the ESB callback */
void esb_event_handler( nrf_esb_evt_t const * p_event ) {
	static nrf_esb_payload_t tmp;
	switch( p_event->evt_id ) {
		case NRF_ESB_EVENT_TX_SUCCESS:
			{
				#if M_DEBUG
					int count = sprintf( DEBUG_MSG, "ESB TX SUCCESS EVENT\r\n" );
					nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
				#endif
			}
			break;
		case NRF_ESB_EVENT_TX_FAILED:
			{
				#if M_DEBUG
					int count = sprintf( DEBUG_MSG, "ESB TX FAILED EVENT\r\n" );
					nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
				#endif
				( void ) nrf_esb_flush_tx();
				( void ) nrf_esb_start_tx();
			}
			break;
		case NRF_ESB_EVENT_RX_RECEIVED:
			{
				while ( nrf_esb_read_rx_payload( &tmp ) == NRF_SUCCESS ) {
					if ( tmp.length > 0 ) {
						memcpy( &rx_payload[ tmp.pipe ], &tmp, sizeof( nrf_esb_payload_t ) );
						received[ tmp.pipe ] = true;
						#if M_DEBUG
							int count = sprintf( DEBUG_MSG, "ESB RX RECEIVED PAYLOAD FROM %02d\r\n", tmp.pipe );
							nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
						#endif
					}
				}
			}
			break;
		default:
			break;
	}
	NRF_GPIO->OUTCLR = 0xFUL << 12;
	NRF_GPIO->OUTSET = ( p_event->tx_attempts & 0x0F ) << 12;
}

/* Request the high frequency clock for the protocol */
void hfclk_init() {
	NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
	NRF_CLOCK->TASKS_HFCLKSTART = 1;
	while ( NRF_CLOCK->EVENTS_HFCLKSTARTED == 0 );
}

/* Initialize and configure the Enhanced ShockBurst protocol */
uint32_t esb_init( void ) {
	uint32_t error;

	// create pipe addresses and prefixes
	uint8_t base_addr0[ 4 ] = { 0xE7, 0xE7, 0xE7, 0xE7 };
	uint8_t base_addr1[ 4 ] = { 0xC2, 0xC2, 0xC2, 0xC2 };
	uint8_t addr_prefix[ 8 ] = { 0xE7, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8 };

	// set transmission protocol
	nrf_esb_config_t config		= NRF_ESB_DEFAULT_CONFIG;
	config.protocol				= NRF_ESB_PROTOCOL_ESB_DPL;
	config.bitrate				= NRF_ESB_BITRATE_2MBPS;
	config.event_handler		= esb_event_handler;
	config.mode					= NRF_ESB_MODE_PRX;
	config.payload_length		= RX_PAYLOAD_LENGTH;
	config.selective_auto_ack	= false;

	error = nrf_esb_init( &config );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_rf_channel( NRF_ESB_CHANNEL );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_base_address_0( base_addr0 );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_base_address_1( base_addr1 );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_prefixes( addr_prefix, 8 );
	VERIFY_SUCCESS( error );

	return error;
}

ret_code_t uart_init( void ) {
	ret_code_t err_code;

	// set UART protocol
	nrf_drv_uart_config_t config = NRF_DRV_UART_DEFAULT_CONFIG;
	config.pseltxd = TX_PIN_NUMBER;
	config.pselrxd = RX_PIN_NUMBER;
	config.pselcts = CTS_PIN_NUMBER;
	config.pselrts = RTS_PIN_NUMBER;
	config.p_context = NULL;
	config.hwfc = NRF_UART_HWFC_DISABLED;
	config.parity = NRF_UART_PARITY_EXCLUDED;
	config.baudrate = NRF_UART_BAUDRATE_115200;
	config.interrupt_priority = UART_DEFAULT_CONFIG_IRQ_PRIORITY;

	// initialize and flush both FIFOs
	err_code = nrf_drv_uart_init( &uart, &config, NULL );

	// assume all peripherals are active until proven otherwise
	for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
		active[ i ] = true;
	}

	return err_code;
}

uint8_t chksum( uint8_t * buf, uint8_t buf_size ) {
	uint16_t sum = 0;
	for ( int i = 0; i < buf_size; i++ ) {
		sum += buf[ i ];
	}
	return ( ( uint8_t ) ( sum % 256 ) );
}

int main( void ) {
	ret_code_t err_code;
	#if M_DEBUG
		int count;
	#endif	 

	bsp_board_leds_init();

	hfclk_init();

	err_code = uart_init();
	APP_ERROR_CHECK( err_code );
	#if M_DEBUG
		count = sprintf( DEBUG_MSG, "UART INTERFACE INITIALIZED\r\n" );
		nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
	#endif

	err_code = esb_init();
	APP_ERROR_CHECK( err_code );
	#if M_DEBUG
		count = sprintf( DEBUG_MSG, "ENHANCED SHOCKBURST INITIALIZED\r\n" );
		nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
	#endif

	err_code = nrf_esb_start_rx();
	APP_ERROR_CHECK( err_code );

	nrf_drv_uart_rx_enable( &uart );

	bsp_board_led_on( 0 );

	uint8_t cmd = 0x00;
	while ( true ) {
		#if M_DEBUG
			count = sprintf( DEBUG_MSG, "RESET REASON: %ld\r\n", NRF_POWER->RESETREAS );
			nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );

			count = sprintf( DEBUG_MSG, "EVENTS RXRDY: %ld\r\n", NRF_UART0->EVENTS_RXDRDY );
			nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
		#endif
		/*
		if ( NRF_POWER->RESETREAS > 0x07 ) { 
			// NRF_POWER->RESETREAS = 0xFFFFFFFF;
			// NRF_POWER->GPREGRET = 0x01;
			NVIC_SystemReset(); 
		}
		*/
		if ( NRF_UART0->EVENTS_RXDRDY ) {		// FIXME: NOT BEING RESET ON POWER ON
			nrf_drv_uart_rx( &uart, &cmd, 1 );
			#if M_DEBUG
				if ( cmd != 0x00 ) {
					count = sprintf( DEBUG_MSG, "RECEIVED COMMAND: %d\r\n", cmd );
					nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count + 1 );
				}
			#endif
			switch ( cmd ) {
				case 0x69: 	// i(nitialize)
					{
						uint8_t init = ( memcmp( active, received, MAX_PERIPHERALS ) ) ? 0x6E : 0x79;
						nrf_drv_uart_tx( &uart, &init, 1 );
					}
					break;
				case 0x063: // c(hannels)
					{
						uint8_t num_sensors = 0;
						uint8_t sensors = 0;
						nrf_drv_uart_rx( &uart, &sensors, 1 );
						uint8_t mask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
						for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
							active[ i ] = ( sensors & mask[ i ] ) ? true : false;
							received[ i ] = false;
							if ( active[ i ] ) {
								#if M_DEBUG
									count = sprintf( DEBUG_MSG, "SENSOR ACTIVATED: %d\n", i );
									nrf_drv_uart_tx( &uart, ( uint8_t * ) DEBUG_MSG, count );
								#endif
								num_sensors++;
							}
						}

						nrf_drv_uart_tx( &uart, &num_sensors, 1 );
					}
					break;
				case 0x77: 	// w(one)
					{
						if ( !memcmp( active, received, MAX_PERIPHERALS ) ) {
							int idx = 0;
							for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
								if ( active[ i ] ) {
									memcpy( &tx_uart_pkt[ idx ], &rx_payload[ i ].data, 4 * sizeof( float ) );
									idx += ( 4 * sizeof( float ) );
								}
							}
							tx_uart_pkt[ idx ] = chksum( tx_uart_pkt, idx );
							nrf_drv_uart_tx( &uart, tx_uart_pkt, idx + 1 );
						}
					}
					break;
				case 0x61:	// a(ccelerometer)
					{
						if ( !memcmp( active, received, MAX_PERIPHERALS ) ) {
							int idx = 0;
							for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
								if ( active[ i ] ) {
									memcpy( &tx_uart_pkt[ idx + 4 * sizeof( float ) ], rx_payload[ i ].data, 3 * sizeof( float ) );
									idx += ( 4 * sizeof( float ) );
								}
							}
							tx_uart_pkt[ idx ] = chksum( tx_uart_pkt, idx );
							nrf_drv_uart_tx( &uart, tx_uart_pkt, idx + 1 );
						}
					}
					break;
				case 0x67:	// g(yroscope)
					{
						if ( !memcmp( active, received, MAX_PERIPHERALS ) ) {
							int idx = 0;
							for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
								if ( active[ i ] ) {
									memcpy( &tx_uart_pkt[ idx + 7 * sizeof( float ) ], rx_payload[ i ].data, 3 * sizeof( float ) );
									idx += ( 4 * sizeof( float ) );
								}
							}
							tx_uart_pkt[ idx ] = chksum( tx_uart_pkt, idx );
							nrf_drv_uart_tx( &uart, tx_uart_pkt, idx + 1 );
						}
					}
					break;
				case 0x6D:	// m(agnetometer)
					{
						if ( !memcmp( active, received, MAX_PERIPHERALS ) ) {
							int idx = 0;
							for ( int i = 0; i < MAX_PERIPHERALS; i++ ) {
								if ( active[ i ] ) {
									memcpy( &tx_uart_pkt[ idx + 10 * sizeof( float ) ], rx_payload[ i ].data, 3 * sizeof( float ) );
									idx += ( 4 * sizeof( float ) );
								}
							}
							tx_uart_pkt[ idx ] = chksum( tx_uart_pkt, idx );
							nrf_drv_uart_tx( &uart, tx_uart_pkt, idx + 1 );
						}
					}
					break;
				default:
					break;
			}
		}
		// nrf_delay_ms( 100 ); // TODO: REMOVE THIS LATER
	}
	return 0;
}



