/*
 * main.c
 *
 *  Created on: Aug 7, 2017
 *      Author: Christopher Hunt <chunt11@jhmi.edu>
 */


// NOTE: TX_PERIPHERAL_NUMBER is the only thing that differs between IMU nodes.
//		 TX_PERIPHERAL_NUMBER may range from 0x00 to 0x07.
#define TX_PERIPHERAL_NUMBER				0x04
#define NRF_LOG_MODULE_NAME					"MAIN"
#define NRF_ESB_CHANNEL						64
#define TX_PAYLOAD_LENGTH					52
#define NUM_CALIB_BYTES						48

#include "common.h"

#include "mpu9250.h"
#include "nrf_esb.h"
#include "nrf_nvmc.h"

// set TX payload to leave from pipe n, with data[ 52 ]
static nrf_esb_payload_t tx_payload = NRF_ESB_CREATE_PAYLOAD( TX_PERIPHERAL_NUMBER,
															  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
															  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
															  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
															  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  );
static nrf_esb_payload_t rx_payload;
volatile bool calibration_flag = false;

/* Event handler for the ESB callback */
void esb_event_handler( nrf_esb_evt_t const * p_event ) {
	switch( p_event->evt_id ) {
		case NRF_ESB_EVENT_TX_SUCCESS:
			NRF_LOG_DEBUG( "TX SUCCESS EVENT\r\n" );
			NRF_LOG_HEXDUMP_DEBUG( tx_payload.data, tx_payload.length );
			break;
		case NRF_ESB_EVENT_TX_FAILED:
			NRF_LOG_DEBUG( "TX FAILED EVENT\r\n" );
			( void ) nrf_esb_flush_tx();
			( void ) nrf_esb_start_tx();
			break;
		case NRF_ESB_EVENT_RX_RECEIVED:
			NRF_LOG_DEBUG( "RX RECEIVED EVENT\r\n" );
			while ( nrf_esb_read_rx_payload( &rx_payload ) == NRF_SUCCESS ) {
				if ( rx_payload.length > 0 ) {
					NRF_LOG_DEBUG( "RX RECEIVED PAYLOAD\r\n" );
				}
			}
			break;
		default:
			break;
	}
	NRF_GPIO->OUTCLR = 0xFUL << 12;
	NRF_GPIO->OUTSET = ( p_event->tx_attempts & 0x0F ) << 12;
}

/* calibration interrupt handler function */
void calibration_interrupt_handler( nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action ) {
	calibration_flag = true;
}

/* Write calibration data to non-volatile memory */
void write_calibration_flash( uint8_t * data, uint32_t data_size ) {
	uint32_t pg_size = NRF_FICR->CODEPAGESIZE;
	uint32_t pg_num = NRF_FICR->CODESIZE - 1;
	uint32_t addr = pg_size * pg_num;

	uint8_t to_write[ data_size + 1 ];
	uint32_t sum = 0;
	for ( int i = 0; i < data_size; i++ ) {
		sum += data[ i ];
	}

	to_write[ data_size ] = sum % 256;
	memcpy( to_write, data, data_size );

	nrf_nvmc_page_erase( addr );
	nrf_nvmc_write_bytes( addr, to_write, sizeof( to_write ) );
}

/* Read calibration from non-volatile memory */
bool read_calibration_flash( uint8_t ** c_data, uint32_t data_size ) {
	// calculate calibration location
	uint32_t pg_size = NRF_FICR->CODEPAGESIZE;
	uint32_t pg_num = NRF_FICR->CODESIZE - 1;
	uint8_t * addr = ( uint8_t * ) ( pg_size * pg_num );

	// check for consistency
	uint32_t sum = 0;
	for ( int i = 0; i < data_size; i++ ) {
		sum += addr[ i ];
	}

	// if consistent, point to the data
	if ( sum && ( ( sum % 256 ) == ( addr[ data_size ] ) ) ) {
		*c_data = addr;
		return true;
	} else { return false; }
}

/* Push button external interrupt */
void gpiote_interrupt_init( void ) {
	ret_code_t err_code;
    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK( err_code );

    nrf_drv_gpiote_in_config_t in_config = GPIOTE_CONFIG_IN_SENSE_HITOLO( true );
    in_config.pull = NRF_GPIO_PIN_PULLUP;

    err_code = nrf_drv_gpiote_in_init( CALIB_PIN, &in_config, calibration_interrupt_handler );
    APP_ERROR_CHECK( err_code );

    nrf_drv_gpiote_in_event_enable( CALIB_PIN, true );
}

/* Request the high frequency clock for the protocol */
void hfclk_init( void ) {
	NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
	NRF_CLOCK->TASKS_HFCLKSTART = 1;
	while ( NRF_CLOCK->EVENTS_HFCLKSTARTED == 0 );
}

/* Initialize and configure the Enhanced ShockBurst protocol */
uint32_t esb_init( void ) {
	uint32_t error;

	// create pipe addresses and prefixes [TODO: VERIFY THESE ARE VALID ADDRESSES]
	uint8_t base_addr0[ 4 ] = { 0xE7, 0xE7, 0xE7, 0xE7 };
	uint8_t base_addr1[ 4 ] = { 0xC2, 0xC2, 0xC2, 0xC2 };
	uint8_t addr_prefix[ 8 ] = { 0xE7, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8 };

	// set transmission protocol
	nrf_esb_config_t config		= NRF_ESB_DEFAULT_CONFIG;
	config.protocol				= NRF_ESB_PROTOCOL_ESB_DPL;
	config.retransmit_delay		= 600;
	config.bitrate				= NRF_ESB_BITRATE_2MBPS;
	config.event_handler		= esb_event_handler;
	config.mode					= NRF_ESB_MODE_PTX;
	config.payload_length		= TX_PAYLOAD_LENGTH;
	config.tx_output_power		= NRF_ESB_TX_POWER_4DBM;
	config.selective_auto_ack	= false;

	error = nrf_esb_init( &config );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_rf_channel( NRF_ESB_CHANNEL );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_base_address_0( base_addr0 );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_base_address_1( base_addr1 );
	VERIFY_SUCCESS( error );

	error = nrf_esb_set_prefixes( addr_prefix, 8 );
	VERIFY_SUCCESS( error );

	return error;
}

int main( void ) {
	ret_code_t err_code;

	bsp_board_leds_init();

	err_code = NRF_LOG_INIT( NULL );
	APP_ERROR_CHECK( err_code );

	NRF_LOG_DEBUG( "INITIALIZING IMU\r\n" );
	if ( init_imu() ) {
		const float * data = NULL;

		NRF_LOG_DEBUG( "READING CALIBRATION DATA FROM FLASH\r\n" );
		if ( read_calibration_flash( ( uint8_t ** ) &data, NUM_CALIB_BYTES ) ) {
			NRF_LOG_DEBUG( "CALIBRATION READ SUCCESSFUL\r\n" );
			NRF_LOG_DEBUG( "ACCEL BIAS: [%d.%d, %d.%d, %d.%d]\r\n", 	NRF_LOG_FLOAT( data[ 0 ] ), NRF_LOG_FLOAT( data[  1 ] ), NRF_LOG_FLOAT( data[  2 ] ) );
			NRF_LOG_DEBUG( "GYRO BIAS: [%d.%d, %d.%d, %d.%d]\r\n",  	NRF_LOG_FLOAT( data[ 3 ] ), NRF_LOG_FLOAT( data[  4 ] ), NRF_LOG_FLOAT( data[  5 ] ) );
			NRF_LOG_DEBUG( "MAG HARD BIAS: [%d.%d, %d.%d, %d.%d]\r\n", 	NRF_LOG_FLOAT( data[ 6 ] ), NRF_LOG_FLOAT( data[  7 ] ), NRF_LOG_FLOAT( data[  8 ] ) );
			NRF_LOG_DEBUG( "MAG SOFT BIAS: [%d.%d, %d.%d, %d.%d]\r\n", 	NRF_LOG_FLOAT( data[ 9 ] ), NRF_LOG_FLOAT( data[ 10 ] ), NRF_LOG_FLOAT( data[ 11 ] ) );
			set_calibration( data );
		} else {
			NRF_LOG_DEBUG( "NO CALIBRATION DATA TO READ\r\n" );
		}

		gpiote_interrupt_init();
		NRF_LOG_DEBUG( "GPIOTE INTERRUPT INITIALIZED\r\n" );

		hfclk_init();
		NRF_LOG_DEBUG( "HIGH FREQUENCY CLOCK INITIALIZED\r\n" );

		err_code = esb_init();
		APP_ERROR_CHECK( err_code );
		NRF_LOG_DEBUG( "ENHANCED SHOCKBURST INITIALIZED\r\n" );

		bsp_board_led_on( 0 );
		NRF_LOG_DEBUG( "MOTION TRACKING TX TRANSMITTING\r\n" );

		while ( true ) {
			// check if someone asked for calibration [don't calibrate during interrupt to avoid race condition]
			if ( calibration_flag ) {
				ret_code_t err;
				NRF_LOG_DEBUG( "CALIBRATION HANDLER ACTIVATED\r\n" );

				// suspend radio activity
				NRF_LOG_DEBUG( "SUSPENDING RADIO ACTIVITY\r\n" );
				err = nrf_esb_suspend();
				APP_ERROR_CHECK( err );

				// flash LED 3x
				for ( int i = 0; i < 3; i++ ) {
				 	bsp_board_led_on( 0 ); nrf_delay_us( 500000 );
				   	bsp_board_led_off( 0 ); nrf_delay_us( 500000 );
				}
				NRF_LOG_DEBUG( "CALIBRATING MPU9250\r\n" );
				calibrate_mpu9250();

				// flash LED 3x
				for ( int i = 0; i < 3; i++ ) {
					bsp_board_led_on( 0 ); nrf_delay_us( 500000 );
				    bsp_board_led_off( 0 ); nrf_delay_us( 500000 );
				}
			    NRF_LOG_DEBUG( "CALIBRATING AK8963\r\n" );
			    calibrate_ak8963();

				// store in persistent memory
			    NRF_LOG_DEBUG( "GET CALIBRATION DATA\r\n" );
			    const float * c_data = get_calibration();

			    NRF_LOG_DEBUG( "WRITING CALIBRATION DATA TO FLASH\r\n" );
			    write_calibration_flash( ( uint8_t * ) c_data, NUM_CALIB_BYTES );

				// restart radio
				NRF_LOG_DEBUG( "RESTARTING RADIO ACTIVITY\r\n" );
				err = esb_init();
				APP_ERROR_CHECK( err );

				// turn on light and reset flag, we're done
				bsp_board_led_on( 0 );
				calibration_flag = false;
			}

			// update data from MPU9250
			update_imu_data();
			compute_orientation();

			// copy data into payload to send
			data = get_orientation();
			memcpy( &tx_payload.data[  0 ], data, 16 );

			data = get_accelerometer();
			memcpy( &tx_payload.data[ 16 ], data, 12 );

			data = get_gyroscope();
			memcpy( &tx_payload.data[ 28 ], data, 12 );

			data = get_magnetometer();
			memcpy( &tx_payload.data[ 40 ], data, 12 );

			tx_payload.noack = false;

			// send data
			if ( nrf_esb_write_payload( &tx_payload ) != NRF_SUCCESS ) {
				NRF_LOG_WARNING( "SENDING PACKET FAILED\r\n" );
			}
			// nrf_delay_us( 2000 ); // wait 2 milliseconds
		}
	} else {
		while( true ) {
			bsp_board_led_invert( 0 );
			NRF_LOG_DEBUG( "UNABLE TO INITIATE IMU\r\n" );
			nrf_delay_ms( 1000 ); // every second
		}
	}
	return 0;
}



