/*
 * MPU9250.c
 *
 *  Created on: Aug 6, 2017
 *      Author: Christopher Hunt <chunt11@jhmi.edu>
 */

#include "mpu9250.h"
#define NRF_LOG_MODULE_NAME	"MPU9250"

// global variables
static float m_orientation[ 4 ] = { 1.0f, 0.0f, 0.0f, 0.0f };
static float m_accel[ 3 ] = { 0.0f, 0.0f, 0.0f };
static float m_gyro[ 3 ] = { 0.0f, 0.0f, 0.0f };
static float m_mag[ 3 ] = { 0.0f, 0.0f, 0.0f };
static float m_temp = 0.0f;
static float m_calibration[ 12 ] = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f };

/* Static declarations for encapsulation. */
static const nrf_drv_twi_t m_twi = NRF_DRV_TWI_INSTANCE( 0 );
static ret_code_t m_err_code;                                 	// so we don't have to keep making a new one

// scales
static float m_accel_scale = 1.0f;
static float m_gyro_scale = 1.0f;
static float m_mag_scale[ 3 ] = { 1.0f, 1.0f, 1.0f };

// feedback variables
static float m_integral_fb[ 3 ] = { 0.0f, 0.0f, 0.0f };

static void rtc_app_timer_timeout_handler( void * p ) {
	// do nothing
}

/* App timer configuration and start */
static void rtc_init( void ) {
	// NRF_LOG_DEBUG( "RTC INITIALIZATION\r\n" );
	if ( !nrf_drv_clock_lfclk_is_running() ) {
		m_err_code = nrf_drv_clock_init();
		APP_ERROR_CHECK( m_err_code );
		nrf_drv_clock_lfclk_request( NULL );
	}

	APP_TIMER_INIT( APP_TIMER_PRESCALER, OP_QUEUES_SIZE, NULL);
	APP_TIMER_DEF( timer_id );

	m_err_code = app_timer_create( &timer_id, APP_TIMER_MODE_REPEATED, rtc_app_timer_timeout_handler );
	APP_ERROR_CHECK( m_err_code );

	m_err_code = app_timer_start( timer_id, APP_TIMER_CLOCK_FREQ, rtc_app_timer_timeout_handler );
	APP_ERROR_CHECK( m_err_code );
}

/* calibrate the accelerometer and gyroscope data */
void calibrate_mpu9250( void ) {
    static const uint16_t sample_count = 128;

    float accel_bias[ 3 ] = { 0.0f, 0.0f, 0.0f };
    float gyro_bias[ 3 ] = { 0.0f, 0.0f, 0.0f };

    // clear previous biases
    // NRF_LOG_DEBUG( "CLEAR BIASES\r\n" );
    for ( int i = 0; i < 3; i++ ) {
    	m_calibration[ i ] = 0.0f;
    	m_calibration[ 3 + i ] = 0.0f;
    }

    // get new stationary samples
    for ( uint16_t i = 0; i < sample_count; i++ ) {
    	// NRF_LOG_DEBUG( "SAMPLING\r\n" );
        update_imu_data();
        for ( int k = 0; k < 3; k++ ) {
            accel_bias[ k ] += m_accel[ k ];
            gyro_bias[ k ] += m_gyro[ k ];
        }
        nrf_delay_us( 135000 );
    }

    // average over sample count
    // NRF_LOG_DEBUG( "CALCULATING\r\n" );
    for ( int i = 0; i < 3; i++ ) {
        accel_bias[ i ] /= ( ( float ) sample_count );
        gyro_bias[ i ] /= ( ( float ) sample_count );
    }
    accel_bias[ 2 ] = fabs( accel_bias[ 2 ] - GRAVITY );

    // store outside of function
    // NRF_LOG_DEBUG( "STORING\r\n" );
    memcpy( &m_calibration[ 0 ], accel_bias, sizeof( accel_bias ) );
    memcpy( &m_calibration[ 3 ], gyro_bias,  sizeof( gyro_bias ) );
}

void calibrate_ak8963( void ) {
    const static uint16_t sample_count = 128;

    float mag_max[ 3 ] = { -1000000.0, -1000000.0, -1000000.0 };
    float mag_min[ 3 ] = {  1000000.0,  1000000.0,  1000000.0 };

    float mag_hard_bias[ 3 ] = { 0.0f, 0.0f, 0.0f };
    float mag_soft_bias[ 3 ] = { 0.0f, 0.0f, 0.0f };

    // NRF_LOG_DEBUG( "CLEAR BIASES\r\n" );
    for ( int i = 0; i < 3; i++ ) {
        m_calibration[ 6 + i ] = 0.0f;
    	m_calibration[ 9 + i ] = 1.0f;
    }

    // collect samples of dynamic magnetometer readings
    for ( uint16_t i = 0; i < sample_count; i++ ) {
    	// NRF_LOG_DEBUG( "SAMPLING\r\n" );
    	update_imu_data();
    	for ( int k = 0; k < 3; k++ ) {
    		mag_hard_bias[ k ] += m_mag[ k ];
    		if ( m_mag[ k ] > mag_max[ k ] ) mag_max[ k ] = m_mag[ k ];
    		if ( m_mag[ k ] < mag_min[ k ] ) mag_min[ k ] = m_mag[ k ];
    }
    	nrf_delay_us( 135000 );
    }

    // calculate biases
    // NRF_LOG_DEBUG( "CALCULATING\r\n" );
    float scales[ 3 ] = { 0.0f, 0.0f, 0.0f };
    float avg_rad = 0.0f;
    for ( int i = 0; i < 3; i++ ) {
        mag_hard_bias[ i ] /= ( ( float ) sample_count );
        scales[ i ] = ( mag_max[ i ] - mag_min[ i ] ) / 2.0f;
        avg_rad += ( scales[ i ] / 3.0f );
    }
    for ( int i = 0; i < 3; i++ ) {
        mag_soft_bias[ i ] = avg_rad / ( ( float ) scales[ i ] );
    }

    // store outside of function
    // NRF_LOG_DEBUG( "STORING\r\n" );
    memcpy( &m_calibration[ 6 ], mag_hard_bias, sizeof( mag_hard_bias ) );
    memcpy( &m_calibration[ 9 ], mag_soft_bias, sizeof( mag_soft_bias ) );
}

const float * get_calibration( void ) {
	return m_calibration;
}

void set_calibration( const float * c_data ) {
	memcpy( m_calibration, c_data, sizeof( m_calibration ) );
}

/* Initialize the IMU */
static void i2c_init( void ) {
	// NRF_LOG_DEBUG( "I2C INITIALIZATION\r\n" );
    // just initialize I2C
    const nrf_drv_twi_config_t config = { .scl                  = I2C_SCL_PIN,
                                          .sda                  = I2C_SDA_PIN,
                                          .frequency            = NRF_TWI_FREQ_400K,
                                          .interrupt_priority   = APP_IRQ_PRIORITY_LOW, // HIGHEST = 0, HIGH = 1, MID = 2, LOW = 3
                                          .clear_bus_init       = false
                                        };
    m_err_code = nrf_drv_twi_init( &m_twi, &config, NULL, NULL );	// blocking mode
    APP_ERROR_CHECK( m_err_code );

    nrf_drv_twi_enable( &m_twi );
}

/* Blocking write call for I2C */
static void i2c_write( uint8_t address, uint8_t * data, int count ) {
	m_err_code = nrf_drv_twi_tx( &m_twi, address, data, count, false );
	if ( m_err_code != NRF_SUCCESS ) {
		NRF_LOG_DEBUG( "I2C WRITE ERROR: %d\r\n", m_err_code );
	}
    APP_ERROR_CHECK( m_err_code );
}

/* Blocking reading call for I2C */
static void i2c_read( uint8_t address, uint8_t * data, int count ) {
    m_err_code = nrf_drv_twi_rx( &m_twi, address, data, count );
    if ( m_err_code != NRF_SUCCESS ) {
    	NRF_LOG_DEBUG( "I2C READ ERROR: %d\r\n", m_err_code );
    }
    APP_ERROR_CHECK( m_err_code );
}

/* Read from the MPU9250 registers */
static void mpu9250_register_read( uint8_t sub_addr, uint8_t count, uint8_t * dest ) {
    uint8_t buff[ 1 ] = { sub_addr };
    i2c_write( MPU9250_ADDR, buff, sizeof( buff ) );
    i2c_read( MPU9250_ADDR, dest, count );
}

/* Write to the MPU9250 register */
static bool mpu9250_register_write( uint8_t sub_addr, uint8_t data ) {
    uint8_t buff[ 2 ] = { sub_addr, data };
    i2c_write( MPU9250_ADDR, buff, sizeof( buff ) );
    mpu9250_register_read( sub_addr, 1, &buff[ 1 ] );
    NRF_LOG_DEBUG( "MPU9250_REGISTER_WRITE %03d [%03d, %03d]: %d\r\n", buff[ 0 ], data, buff[ 1 ], buff[ 1 ] == data );
    return ( buff[ 1 ] == data );
}

/* Check the WHO_AM_I register of the MPU9250 */
static bool mpu9250_who_am_i() {
    uint8_t buff[ 1 ];
    mpu9250_register_read( WHO_AM_I, sizeof( buff ), &buff[ 0 ] );
    return ( buff[ 0 ] == 0x71 );
}

/* Read from the AK8963 register */
static void ak8963_register_read( uint8_t sub_addr, uint8_t count, uint8_t * dest ) {
    mpu9250_register_write( I2C_SLV0_ADDR, AK8963_I2C_ADDR | I2C_READ_FLAG );
    mpu9250_register_write( I2C_SLV0_REG, sub_addr );
    mpu9250_register_write( I2C_SLV0_CTRL, I2C_SLV0_EN | count );
    nrf_delay_us( 100 );
    mpu9250_register_read( EXT_SENS_DATA_00, count, dest );
}

/* Write to the AK8963 register */
static bool ak8963_register_write( uint8_t sub_addr, uint8_t data ) {
    uint8_t count = 1;
    uint8_t buff[ 1 ];

    mpu9250_register_write( I2C_SLV0_ADDR, AK8963_I2C_ADDR );
    mpu9250_register_write( I2C_SLV0_REG, sub_addr );
    mpu9250_register_write( I2C_SLV0_DO, data );
    mpu9250_register_write( I2C_SLV0_CTRL, I2C_SLV0_EN | count );

    ak8963_register_read( sub_addr, sizeof( buff ), &buff[ 0 ] );
    NRF_LOG_DEBUG( "AK8963_REGISTER_WRITE %03d [%03d, %03d]: %d\r\n", sub_addr, data, buff[ 0 ], buff[ 0 ] == data );
    return ( buff[ 0 ] == data );
}

/* Check the WHO_AM_I register of the MPU9250 */
static bool ak8963_who_am_i() {
    uint8_t buff[ 1 ];
    ak8963_register_read( AK8963_WHO_AM_I, sizeof( buff ), &buff[ 0 ] );
    return ( buff[ 0 ] == 0x48 );
}


/* Set the accelerometer measurement range and change the scaling accordingly */
static bool set_accel_scale( accel_range_t range ) {
    static const float ACCEL_SCALE_FACTOR = GRAVITY / 32767.5f;
    switch( range ) {
        case ACCEL_RANGE_2G:
            if ( mpu9250_register_write( ACCEL_CONFIG, ACCEL_FS_SEL_2G ) ) {
                m_accel_scale = 2.0f * ACCEL_SCALE_FACTOR;
                return true;
            }
            break;
        case ACCEL_RANGE_4G:
            if ( mpu9250_register_write( ACCEL_CONFIG, ACCEL_FS_SEL_4G ) ) {
                m_accel_scale = 4.0f * ACCEL_SCALE_FACTOR;
                  return true;
            }
            break;
        case ACCEL_RANGE_8G:
            if ( mpu9250_register_write( ACCEL_CONFIG, ACCEL_FS_SEL_8G ) ) {
                m_accel_scale = 8.0f * ACCEL_SCALE_FACTOR;
                return true;
            }
            break;
        case ACCEL_RANGE_16G:
            if ( mpu9250_register_write( ACCEL_CONFIG, ACCEL_FS_SEL_16G ) ) {
                m_accel_scale = 16.0f * ACCEL_SCALE_FACTOR;
                return true;
            }
            break;
        default:
            break;
    }
    return false;
}

/* Set the gyroscope measuring range and change the scale factor accordingly */
static bool set_gyro_scale( gyro_range_t range ) {
    static const float GYRO_SCALE_FACTOR = DEG_2_RAD / 32767.5f;
    switch( range ) {
        case GYRO_RANGE_250DPS:
            if ( mpu9250_register_write( GYRO_CONFIG, GYRO_FS_SEL_250DPS ) ) {
                m_gyro_scale = 250.0f * GYRO_SCALE_FACTOR;
                return true;
            }
            break;
        case GYRO_RANGE_500DPS:
            if ( mpu9250_register_write( GYRO_CONFIG, GYRO_FS_SEL_500DPS ) ) {
                m_gyro_scale = 500.0f * GYRO_SCALE_FACTOR;
                return true;
            }
            break;
        case GYRO_RANGE_1000DPS:
            if ( mpu9250_register_write( GYRO_CONFIG, GYRO_FS_SEL_1000DPS ) ) {
                m_gyro_scale = 1000.0f * GYRO_SCALE_FACTOR;
                return true;
            }
            break;
        case GYRO_RANGE_2000DPS:
            if ( mpu9250_register_write( GYRO_CONFIG, GYRO_FS_SEL_2000DPS ) ) {
                m_gyro_scale = 2000.0f * GYRO_SCALE_FACTOR;
                return true;
            }
            break;
        default:
            break;
    }
    return false;
}

/* Set the magnetometer measurement precision and change the scale accordingly */
static bool set_mag_scale( mag_range_t range ) {
    uint8_t buff[ 3 ];
    if ( ak8963_register_write( AK8963_CNTL1, AK8963_FUSE_ROM ) ) {
        nrf_delay_us( 10000 );
        ak8963_register_read( AK8963_ASA, sizeof( buff ), &buff[ 0 ] );

        //nrf_delay_ms( 1000 );
        //NRF_LOG_DEBUG( "%d %d %d\r\n", buff[0],buff[1],buff[2] );

        float mag_res = ( range == MFS_16BITS ) ? ( 4912.0f / 32760.0f ) : ( 4912.0f / 8190.0f );
        m_mag_scale[ 0 ] = ( ( ( ( float ) buff[ 0 ] ) - 128.0f ) / 256.0f + 1.0f ) * mag_res;
        m_mag_scale[ 1 ] = ( ( ( ( float ) buff[ 1 ] ) - 128.0f ) / 256.0f + 1.0f ) * mag_res;
        m_mag_scale[ 2 ] = ( ( ( ( float ) buff[ 2 ] ) - 128.0f ) / 256.0f + 1.0f ) * mag_res;
        return true;
    }
    return false;
}

/* Set the register interrupt status */
static bool enable_interrupt( bool enable ) {
    if ( enable ) {
        if ( !mpu9250_register_write( INT_PIN_CFG, INT_PULSE_50US ) ) { return false; }
        if ( !mpu9250_register_write( INT_ENABLE,  INT_RAW_RDY_EN ) ) { return false; }
    } else {
        if ( !mpu9250_register_write( INT_ENABLE, INT_DISABLE ) ) { return false; }
    }
    return true;
}

/* Configure the digital low pass filter. */
static bool set_dlpf( dlpf_bandwidth_t bd, uint8_t srd ) {
    uint8_t buff[ 7 ];
    switch( bd ) {
        case DLPF_BANDWIDTH_184HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_184 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_184 ) ) { return false; }
            break;
        case DLPF_BANDWIDTH_92HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_92 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_92 ) ) { return false; }
            break;
        case DLPF_BANDWIDTH_41HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_41 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_41 ) ) { return false; }
            break;
        case DLPF_BANDWIDTH_20HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_20 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_20 ) ) { return false; }
            break;
        case DLPF_BANDWIDTH_10HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_10 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_10 ) ) { return false; }
            break;
        case DLPF_BANDWIDTH_5HZ:
            if ( !mpu9250_register_write( ACCEL_CONFIG2, ACCEL_DLPF_5 ) ) { return false; }
            if ( !mpu9250_register_write( CONFIG, GYRO_DLPF_5 ) ) { return false; }
            break;
        default:
            return false;
    }
    // set the sample rate divider
    if ( !mpu9250_register_write( SMPDIV, srd ) ) { return false; }
    if ( srd > 9 ) {
        if ( !ak8963_register_write( AK8963_CNTL1, AK8963_PWR_DOWN ) ) { return false; }
        nrf_delay_us( 100000 );
        if ( !ak8963_register_write( AK8963_CNTL1, AK8963_CNT_MEAS1 ) ) { return false; }
        nrf_delay_us( 100000 );
        ak8963_register_read( AK8963_HXL, sizeof( buff ), &buff[ 0 ] );
    }

    // set the register interrupt
    if ( !enable_interrupt( true ) ) { return false; }
    return true;
}


/* Set the MPU9250 parameters */
static bool set_mpu9250_params( void ) {
    uint8_t buff[ 7 ];

    NRF_LOG_DEBUG("SETTING MPU9250 PARAMETERS\r\n")
    if ( !mpu9250_register_write( PWR_MGMNT_1, CLOCK_SEL_PLL ) ) { return false; }			// select clock source to gyro
    if ( !mpu9250_register_write( USER_CTRL, I2C_MST_EN ) ) { return false; }				// enable I2C master mode
    if ( !mpu9250_register_write( I2C_MST_CTRL, I2C_MST_CLK ) ) { return false; }			// set the I2C bus speed to 400 kHz
    if ( !ak8963_register_write( AK8963_CNTL1, AK8963_PWR_DOWN ) ) { return false; }		// set the AK8963 to power down

    mpu9250_register_write( PWR_MGMNT_1, PWR_RESET ); nrf_delay_us( 1000 );					// reset the MPU9250
    ak8963_register_write( AK8963_CNTL2, AK8963_RESET );									// reset the AK8963

    if ( !mpu9250_register_write( PWR_MGMNT_1, CLOCK_SEL_PLL ) ) { return false; }			// select clock source to gyro
    if ( !mpu9250_who_am_i() ) { return false; }											// check the WHO_AM_I byte
    if ( !mpu9250_register_write( PWR_MGMNT_2, SEN_ENABLE ) ) { return false; }				// enable accelerometer and gyroscope

    if ( !set_accel_scale( ACCEL_RANGE_8G ) ) { return false; }								// set accelerometer range
    if ( !set_gyro_scale( GYRO_RANGE_1000DPS ) ) { return false; }							// set gyroscope range

    if ( !mpu9250_register_write( USER_CTRL, I2C_MST_EN ) ) { return false; }				// enable the I2C master mode
    if ( !mpu9250_register_write( I2C_MST_CTRL, I2C_MST_CLK ) ) { return false; }			// set the I2C bus speed to 400 kHz
    if ( !ak8963_who_am_i() ) { return false; }												// check the AK8963_WHO_AM_I byte

    if ( !ak8963_register_write( AK8963_CNTL1, AK8963_PWR_DOWN ) ) { return false; }		// set AK8963 to power down
    nrf_delay_us( 100000 );																	// wait
    if ( !set_mag_scale( MFS_16BITS ) ) { return false; }									// set magnetometer scale with factory calibration
    if ( !ak8963_register_write( AK8963_CNTL1, AK8963_PWR_DOWN ) ) {  return false; }		// set AK8963 to power down
    nrf_delay_us( 100000 );																	// wait
    if ( !ak8963_register_write( AK8963_CNTL1, AK8963_CNT_MEAS2 ) ) {  return false; }		// set AK8963 to 16-bit resolution, 100 Hz update rate
    nrf_delay_us( 100000 );																	// wait

    if ( !mpu9250_register_write( PWR_MGMNT_1, CLOCK_SEL_PLL ) ) {  return false; }			// select clock source to gyro
    ak8963_register_read( AK8963_HXL, sizeof( buff ), &buff[ 0 ] );							// read 7 bytes of data from the AK8963 at specified rate

    if ( !set_dlpf( DLPF_BANDWIDTH_41HZ, 0x04 ) ) { return false; }							// set DLPF to 41 Hz cutoff, using 200 Hz sampling rate

    // NRF_LOG_HEXDUMP_DEBUG( buff, sizeof( buff ) );

    return true;
}

 /* Call this for your initialization needs. */
bool init_imu( void ) {
    NRF_LOG_INIT( NULL );

    nrf_gpio_cfg_input( INT_PIN, NRF_GPIO_PIN_NOPULL );

    rtc_init();
    i2c_init();

    return set_mpu9250_params();
}

/* Call this to update the sensor values */
void update_imu_data( void ) {
    static const float T[ 3 ][ 3 ] = { { 0.0f, 1.0f,  0.0f },
                                       { 1.0f, 0.0f,  0.0f },
                                       { 0.0f, 0.0f, -1.0f } };
    uint8_t buff[ 21 ];
    int16_t ax, ay, az, gx, gy, gz, mx, my, mz, tt;

    mpu9250_register_read( ACCEL_OUT, sizeof( buff ), &buff[ 0 ] );

    // get counts
    ax = ( ( ( int16_t ) buff[ 0 ] ) << 8 ) | buff[ 1 ];
    ay = ( ( ( int16_t ) buff[ 2 ] ) << 8 ) | buff[ 3 ];
    az = ( ( ( int16_t ) buff[ 4 ] ) << 8 ) | buff[ 5 ];

    tt = ( ( ( int16_t ) buff[ 6 ] ) << 8 ) | buff[ 7 ];

    gx = ( ( ( int16_t ) buff[ 8 ] ) << 8 ) | buff[ 9 ];
    gy = ( ( ( int16_t ) buff[ 10 ] ) << 8 ) | buff[ 11 ];
    gz = ( ( ( int16_t ) buff[ 12 ] ) << 8 ) | buff[ 13 ];

    if ( buff[ 20 ] == 0x10 ) {
        mx = ( ( ( int16_t ) buff[ 15 ] ) << 8 ) | buff[ 14 ];
        my = ( ( ( int16_t ) buff[ 17 ] ) << 8 ) | buff[ 16 ];
        mz = ( ( ( int16_t ) buff[ 19 ] ) << 8 ) | buff[ 18 ];
    } else { mx = 0; my = 0;  mz = 0; }

    // get proper units
    m_accel[ 0 ] = ( ( float ) ( T[ 0 ][ 0 ] * ax + T[ 0 ][ 1 ] * ay + T[ 0 ][ 2 ] * az ) ) * m_accel_scale - m_calibration[ 0 ];
    m_accel[ 1 ] = ( ( float ) ( T[ 1 ][ 0 ] * ax + T[ 1 ][ 1 ] * ay + T[ 1 ][ 2 ] * az ) ) * m_accel_scale - m_calibration[ 1 ];
    m_accel[ 2 ] = ( ( float ) ( T[ 2 ][ 0 ] * ax + T[ 2 ][ 1 ] * ay + T[ 2 ][ 2 ] * az ) ) * m_accel_scale - m_calibration[ 2 ];

    m_gyro[ 0 ] = ( ( float ) ( T[ 0 ][ 0 ] * gx + T[ 0 ][ 1 ] * gy + T[ 0 ][ 2 ] * gz ) ) * m_gyro_scale - m_calibration[ 3 ];
    m_gyro[ 1 ] = ( ( float ) ( T[ 1 ][ 0 ] * gx + T[ 1 ][ 1 ] * gy + T[ 1 ][ 2 ] * gz ) ) * m_gyro_scale - m_calibration[ 4 ];
    m_gyro[ 2 ] = ( ( float ) ( T[ 2 ][ 0 ] * gx + T[ 2 ][ 1 ] * gy + T[ 2 ][ 2 ] * gz ) ) * m_gyro_scale - m_calibration[ 5 ];

    m_mag[ 0 ] = ( ( ( float ) mx ) * m_mag_scale[ 0 ] - m_calibration[ 6 ] ) * m_calibration[  9 ];
    m_mag[ 1 ] = ( ( ( float ) my ) * m_mag_scale[ 1 ] - m_calibration[ 7 ] ) * m_calibration[ 10 ];
    m_mag[ 2 ] = ( ( ( float ) mz ) * m_mag_scale[ 2 ] - m_calibration[ 8 ] ) * m_calibration[ 11 ];

    m_temp = ( ( ( float ) tt ) / 333.87 ) + 21.0f;

    /*
    if ( buff[ 20 ] == 0x10 ) {
    	NRF_LOG_DEBUG( "[ %d.%d, %d.%d, %d.%d ]\r\n", NRF_LOG_FLOAT( m_mag[ 0 ] ), NRF_LOG_FLOAT( m_mag[ 1 ] ), NRF_LOG_FLOAT( m_mag[ 2 ] ) );
    } else { NRF_LOG_DEBUG( "OVERFLOW!!!\r\n"); }
    */
}

const float * get_accelerometer( void ) {
	return m_accel;
}

const float * get_gyroscope( void ) {
	return m_gyro;
}

const float * get_magnetometer( void ) {
	return m_mag;
}

const float get_temperature( void ) {
	return m_temp;
}

/* Fast algorithm for approximating the inverse square root of a float */
static float invsqrt( float x ) {
    float halfx = 0.5f * x;
    float y = x;
    long i = *( long* ) &y;
    i = 0x5f3759df - ( i >> 1 );
    y = *( float * ) &i;
    y = y * ( 1.5f - ( halfx * y * y ) );
    y = y * ( 1.5f - ( halfx * y * y ) );
    return y;
}

/* Computes orientation using the Mahony complementary filter */
void compute_orientation( void ) {
    const static float TWO_KP = 2.0f * 10.0f;
    const static float TWO_KI = 2.0f * 0.02f;
    float VALID_GYRO = m_gyro_scale * 32767.5f;

    float norm;
    float ax, ay, az, gx, gy, gz, mx, my, mz;
    float q0q0, q0q1, q0q2, q0q3, q1q1, q1q2, q1q3, q2q2, q2q3, q3q3;
    float hx, hy, bx, bz;
    float halfvx, halfvy, halfvz, halfwx, halfwy, halfwz;
    float halfex, halfey, halfez;
    float qa, qb, qc;

    static uint32_t tick1, tick2;
    static uint32_t dtick = 0;

    // update data and time
    if ( dtick == 0 ) {
    	tick1 = app_timer_cnt_get();																				// initialize tick1
    }
    tick2 = app_timer_cnt_get();																					// get current tick count
    app_timer_cnt_diff_compute( tick2, tick1, &dtick );																// get difference since last tick count [use this to be safe from overflow]
    float deltat = ( ( float ) dtick * ( 1.0f + APP_TIMER_PRESCALER ) ) / APP_TIMER_CLOCK_FREQ;							// convert ticks to seconds
    // NRF_LOG_DEBUG( "DELTAT: %d.%d (%d TICKS)\r\n", NRF_LOG_FLOAT( deltat ), dtick );
    tick1 = tick2; 																									// update timer ticks

    if ( ( fabs( m_gyro[ 0 ] ) < VALID_GYRO ) && ( fabs( m_gyro[ 1 ] ) < VALID_GYRO ) && ( fabs( m_gyro[ 2 ] ) < VALID_GYRO ) ) {
        // store gyroscope readings as rad / sec
        gx = m_gyro[0];
        gy = m_gyro[1];
        gz = m_gyro[2];

        // auxiliary variables to avoid repeated math
        q0q0 = m_orientation[0] * m_orientation[0];
        q0q1 = m_orientation[0] * m_orientation[1];
        q0q2 = m_orientation[0] * m_orientation[2];
        q0q3 = m_orientation[0] * m_orientation[3];
        q1q1 = m_orientation[1] * m_orientation[1];
        q1q2 = m_orientation[1] * m_orientation[2];
        q1q3 = m_orientation[1] * m_orientation[3];
        q2q2 = m_orientation[2] * m_orientation[2];
        q2q3 = m_orientation[2] * m_orientation[3];
        q3q3 = m_orientation[3] * m_orientation[3];

        // initialize correction vectors as zero
        halfex = 0.0f;
        halfey = 0.0f;
        halfez = 0.0f;

        // compute feedback only if accelerometer measurement is valid
        if ( ( m_accel[0] != 0.0f ) || ( m_accel[1] != 0.0f ) || ( m_accel[2] != 0.0f ) ) {
            // normalize accelerometer measurement
            norm = invsqrt( m_accel[0] * m_accel[ 0 ] + m_accel[1] * m_accel[1] + m_accel[2] * m_accel[2] );
            ax = m_accel[0] * norm;
            ay = m_accel[1] * norm;
            az = m_accel[2] * norm;

            // estimated direction of gravity and magnetic field
            halfvx = q1q3 - q0q2;
            halfvy = q0q1 + q2q3;
            halfvz = q0q0 - 0.5f + q3q3;

            if ( ( m_mag[0] != 0.0f ) || ( m_mag[1] != 0.0f ) || ( m_mag[2] != 0.0f ) ) { // valid magnetometer readings
                // normalize magnetometer measurement
                norm = invsqrt( m_mag[0] * m_mag[0] + m_mag[1] * m_mag[1] + m_mag[2] * m_mag[2] );
                mx = m_mag[0] * norm;
                my = m_mag[1] * norm;
                mz = m_mag[2] * norm;
            } else {
            	mx = 0.0f; my = 0.0f; mz = 0.0f;
            }

            // reference direction of Earth's magnetic field
            hx = 2.0f * (mx * (0.5f - q2q2 - q3q3) + my * (q1q2 - q0q3) + mz * (q1q3 + q0q2));
            hy = 2.0f * (mx * (q1q2 + q0q3) + my * (0.5f - q1q1 - q3q3) + mz * (q2q3 - q0q1));
            bx = sqrtf(hx * hx + hy * hy);
            bz = 2.0f * (mx * (q1q3 - q0q2) + my * (q2q3 + q0q1) + mz * (0.5f - q1q1 - q2q2));

            // estimated direction of gravity and magnetic field
            halfwx = bx * (0.5f - q2q2 - q3q3) + bz * (q1q3 - q0q2);
            halfwy = bx * (q1q2 - q0q3) + bz * (q0q1 + q2q3);
            halfwz = bx * (q0q2 + q1q3) + bz * (0.5f - q1q1 - q2q2);

            // error is sum of cross product between estimated direction and measured direction of field vectors
            halfex = (ay * halfvz - az * halfvy) + (my * halfwz - mz * halfwy);
            halfey = (az * halfvx - ax * halfvz) + (mz * halfwx - mx * halfwz);
            halfez = (ax * halfvy - ay * halfvx) + (mx * halfwy - my * halfwx);
        }

        // if there is a valid correction vector
        if ( ( halfex != 0.0f ) || ( halfey != 0.0f ) || ( halfez != 0.0f ) ) {
            // compute and apply integral feedback if enabled
            if ( TWO_KI > 0.0f ) {
                // integral error scaled by Ki
                m_integral_fb[0] += ( TWO_KI * halfex * deltat );
                m_integral_fb[1] += ( TWO_KI * halfey * deltat );
                m_integral_fb[2] += ( TWO_KI * halfez * deltat );

                // apply integral feedback
                gx += m_integral_fb[0];
                gy += m_integral_fb[1];
                gz += m_integral_fb[2];
            } else { // prevent integral windup
                m_integral_fb[0] = 0.0f;
                m_integral_fb[1] = 0.0f;
                m_integral_fb[2] = 0.0f;
            }

            // apply proportional feedback
            gx += ( TWO_KP * halfex );
            gy += ( TWO_KP * halfey );
            gz += ( TWO_KP * halfez );
        }

        // integrate rate of change of quaternion
        gx *= ( 0.5f * deltat );
        gy *= ( 0.5f * deltat );
        gz *= ( 0.5f * deltat );

        qa = m_orientation[0]; qb = m_orientation[1]; qc = m_orientation[2];
        m_orientation[0] += ( -qb * gx - qc * gy - m_orientation[3] * gz );
        m_orientation[1] += (  qa * gx + qc * gz - m_orientation[3] * gy );
        m_orientation[2] += (  qa * gy - qb * gz + m_orientation[3] * gx );
        m_orientation[3] += (  qa * gz + qb * gy - qc * gx );

        norm = invsqrt( m_orientation[0] * m_orientation[0] +
                        m_orientation[1] * m_orientation[1] +
                        m_orientation[2] * m_orientation[2] +
                        m_orientation[3] * m_orientation[3] );
        m_orientation[0] *= norm;
        m_orientation[1] *= norm;
        m_orientation[2] *= norm;
        m_orientation[3] *= norm;

        /*
        NRF_LOG_DEBUG( "%02d: %d.%d\r\n", 1, NRF_LOG_FLOAT( m_orientation[ 0 ] ) );
        NRF_LOG_DEBUG( "%02d: %d.%d\r\n", 2, NRF_LOG_FLOAT( m_orientation[ 1 ] ) );
        NRF_LOG_DEBUG( "%02d: %d.%d\r\n", 3, NRF_LOG_FLOAT( m_orientation[ 2 ] ) );
        NRF_LOG_DEBUG( "%02d: %d.%d\r\n", 4, NRF_LOG_FLOAT( m_orientation[ 3 ] ) );
    	*/
    }
}

const float * get_orientation( void ) {
	return m_orientation;
}

/* Computes the position through double integration */
/*
void compute_position( void ) {
      float acc[ 3 ] = { 0.0f, 0.0f, 0.0f };
      float vel[ 3 ] = { 0.0f, 0.0f, 0.0f };
      float T = { { 1.0f - 2.0f * m_orientation[ 2 ] * m_orientation[ 2 ] - 2.0f * m_orientation[ 3 ] * m_orientation[ 3 ],
                    2.0f * m_orientation[ 1 ] * m_orientation[ 2 ] - 2.0f * m_orientation[ 3 ] * m_orientation[ 0 ],
                    2.0f * m_orientation[ 1 ] * m_orientation[ 3 ] + 2.0f * m_orientation[ 2 ] * m_orientation[ 0 ] },
                  { 2.0f * m_orientation[ 1 ] * m_orientation[ 2 ] + 2.0f * m_orientation[ 3 ] * m_orientation[ 0 ],
                    1.0f - 2.0f * m_orientation[ 1 ] * m_orientation[ 1 ] - 2.0f * m_orientation[ 3 ] * m_orientation[ 3 ],
                    2.0f * m_orientation[ 2 ] * m_orientation[ 3 ] - 2.0f * m_orientation[ 1 ] * m_orientation[ 0 ] },
                  { 2.0f * m_orientation[ 1 ] * m_orientation[ 3 ] - 2.0f * m_orientation[ 2 ] * m_orientation[ 0 ],
                    2.0f * m_orientation[ 2 ] * m_orientation[ 3 ] + 2.0f * m_orientation[ 1 ] * m_orientation[ 0 ],
                    1.0f - 2.0f * m_orientation[ 1 ] * m_orientation[ 1 ] - 2.0f * m_orientation[ 2 ] * m_orientation[ 2 ] } };
      uint32_t dtick;
      static uint32_t tick1 = 0;
      static uint32_t tick2 = 0;

      // update data and time
      if ( tick1 == 0 && tick2 == 0 ) {
    	  app_timer_cnt_get( &tick1 );
      }
      app_timer_cnt_get( &tick2 );										// get current tick count
      app_timer_cnt_diff_compute( tick1, tick2, &dtick );				// get difference since last tick count [use this to be safe from overflow]
      float deltat = ( ( float ) dtick ) / APP_TIMER_CLOCK_FREQ;		// convert ticks to seconds
      tick1 = tick2; 													// update timer ticks

      // get accelerations in correctly-oriented frame and gravity-compensated
      for ( int i = 0; i < 3; i++ ) {
          acc[ i ] = T[ i ][ 0 ] * m_accel[ 0 ] + T[ i ][ 1 ] * m_accel[ 1 ] + T[ i ][ 2 ] * m_accel[ 2 ];
          if ( i == 2 ) { acc[ i ] -= GRAVITY; }
      }

      // double integration to get position
      for ( int i = 0; i < 3; i++ ) {
          vel[ i ] += ( deltat * acc[ i ] );
          m_position[ i ] += ( deltat * vel[ i ] );
      }
}
*/
